# Generated by Django 4.2.6 on 2023-10-18 09:37

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='DataChange',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='FunctionCall',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='RequestCall',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='RequestUrl',
            fields=[
                ('id', models.CharField(max_length=255, primary_key=True, serialize=False)),
                ('is_active', models.BooleanField(default=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ServiceFunction',
            fields=[
                ('id', models.CharField(max_length=255, primary_key=True, serialize=False)),
                ('is_active', models.BooleanField(default=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='FunctionCallArgs',
            fields=[
                ('value', models.JSONField(default=dict)),
                ('anchor', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, primary_key=True, related_name='function_args', serialize=False, to='metadata.functioncall')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='RequestCallArgs',
            fields=[
                ('value', models.JSONField(default=dict)),
                ('anchor', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, primary_key=True, related_name='request_args', serialize=False, to='metadata.requestcall')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='RequestUser',
            fields=[
                ('anchor', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, primary_key=True, related_name='request_user', serialize=False, to='metadata.requestcall')),
                ('value', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='RequestCallUrl',
            fields=[
                ('anchor', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, primary_key=True, related_name='request_url', serialize=False, to='metadata.requestcall')),
                ('value', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='metadata.requesturl')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='FunctionCallServiceFunction',
            fields=[
                ('anchor', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, primary_key=True, related_name='function_signature', serialize=False, to='metadata.functioncall')),
                ('value', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='metadata.servicefunction')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]
